! Program created by Gustavo Catao Alves at
! the Stanford Exploration Project
! May, 6th, 2014

! finite difference module
module elasticlib_mod_new

  use sep

  implicit none
  integer, private   :: nx, nz, sp, iz, ix
  real, private      :: c1,c2,c3,c4,c5

  contains

  ! Initialization of all variables needed by this subroutine
  subroutine elasticmod_init(nz_in,nx_in, sp_in)
    integer        :: nx_in, nz_in, sp_in

    nx = nx_in
    nz = nz_in
    sp = sp_in

  endsubroutine

  ! Functions that are called by elastic_mod

  ! These x derivatives are centered at x=+1/2 
  function dx_op(adj,add,model,data) result(stat)
    logical,intent(in),optional :: adj,add
    real,dimension(:,:)         :: model,data
    integer                     :: stat

    if (sp.eq.4) then
      call dx4th_op2(adj,add,model,data)
    else
      call dx10th_op2(adj,add,model,data)
    endif
    stat=0
  endfunction

  ! These x derivatives are centered at x=-1/2 
  function dx_op3(adj,add,model,data) result(stat)
    logical,intent(in),optional :: adj,add
    real,dimension(:,:)         :: model,data
    integer                     :: stat

    if (sp.eq.4) then
      call dx4th_op4(adj,add,model,data)
    else
      call dx10th_op4(adj,add,model,data)
    endif
    stat=0
  endfunction

  ! These z derivatives are centered at z=-1/2 
  function dz_op(adj,add,model,data) result(stat)
    logical,intent(in),optional :: adj,add
    real,dimension(:,:)         :: model,data
    integer                     :: stat

    if (sp.eq.4) then
      call dz4th_op2(adj,add,model,data)
    else
      call dz10th_op2(adj,add,model,data)
    endif
    stat=0
  endfunction

  ! These z derivatives are centered at z=+1/2 
  function dz_op3(adj,add,model,data) result(stat)
    logical,intent(in),optional :: adj,add
    real,dimension(:,:)         :: model,data
    integer                     :: stat

    if (sp.eq.4) then
      call dz4th_op4(adj,add,model,data)
    else
      call dz10th_op4(adj,add,model,data)
    endif
    stat=0
  endfunction

  ! Derivative used on the shear stress
  function dxz_op(adj,add,model,data1,data2) result(stat)
    logical,intent(in),optional :: adj,add
    real,dimension(:,:)         :: model,data1,data2
    integer                     :: stat

    if (sp.eq.4) then
      call dxz4th_op2(adj,add,model,data1,data2)
    else
      call dxz10th_op2(adj,add,model,data1,data2)
    endif
    stat=0
  endfunction

  ! Subroutines that are called by the functions
  ! d/dx subroutines
  subroutine dx4th_op2(adj,add,model,data)
    logical,optional            :: add,adj
    real,dimension(nz,nx)       :: model,data

    c1=-1.0/24.0
    c2= 9.0/8.0
    model=0.0
    do ix=3,nx-2
      do iz=3,nz-2
        model(iz,ix) = model(iz,ix) + c1*(data(iz,ix+1)-data(iz,ix-2))+&
                                          c2*(data(iz,ix  )-data(iz,ix-1))
      enddo
    enddo
  endsubroutine

  subroutine dx10th_op2(adj,add,model,data)
    logical,optional            :: add,adj
    real,dimension(nz,nx)     :: model,data

    c1=   35.0/294912.0
    c2= -405.0/229376.0
    c3=   567.0/40960.0
    c4=   -735.0/8192.0
    c5= 19845.0/16384.0
    model=0.0
!$OMP PARALLEL DO
    do ix=6,nx-5
      do iz=1,nz-5
        model(iz,ix) = model(iz,ix) + c1*(data(iz,ix+4)-data(iz,ix-5))+&
                                          c2*(data(iz,ix+3)-data(iz,ix-4))+&
                                          c3*(data(iz,ix+2)-data(iz,ix-3))+&
                                          c4*(data(iz,ix+1)-data(iz,ix-2))+&
                                          c5*(data(iz,ix  )-data(iz,ix-1))
      enddo
    enddo
!$OMP END PARALLEL DO
  endsubroutine

  subroutine dx4th_op4(adj,add,model,data)
    logical,optional            :: add,adj
    real,dimension(nz,nx)     :: model,data

    c1=-1.0/24.0
    c2= 9.0/8.0
    model=0.0
!$OMP PARALLEL DO
    do ix=3,nx-2
      do iz=3,nz-2
        model(iz,ix) = model(iz,ix) + c1*(data(iz,ix+2)-data(iz,ix-1))+&
                                          c2*(data(iz,ix+1)-data(iz,ix  ))
      enddo
    enddo
!$OMP END PARALLEL DO
  endsubroutine

  subroutine dx10th_op4(adj,add,model,data)
    logical,optional            :: add,adj
    real,dimension(nz,nx)     :: model,data

    c1=   35.0/294912.0
    c2= -405.0/229376.0
    c3=   567.0/40960.0
    c4=   -735.0/8192.0
    c5= 19845.0/16384.0
    model=0.0
!$OMP PARALLEL DO
    do ix=6,nx-5
      do iz=1,nz-5
        model(iz,ix) = model(iz,ix) + c1*(data(iz,ix+5)-data(iz,ix-4))+&
                                          c2*(data(iz,ix+4)-data(iz,ix-3))+&
                                          c3*(data(iz,ix+3)-data(iz,ix-2))+&
                                          c4*(data(iz,ix+2)-data(iz,ix-1))+&
                                          c5*(data(iz,ix+1)-data(iz,ix  ))
      enddo
    enddo
!$OMP END PARALLEL DO
  endsubroutine

  ! d/dz subroutines
  subroutine dz4th_op2(adj,add,model,data)
    logical,optional            :: add,adj
    real,dimension(nz,nx)     :: model,data

    c1=-1.0/24.0
    c2= 9.0/8.0
    model=0.0
!$OMP PARALLEL DO
    do ix=3,nx-2
      do iz=3,nz-2
        model(iz,ix) = model(iz,ix) + c1*(data(iz+1,ix)-data(iz-2,ix))+&
                                          c2*(data(iz  ,ix)-data(iz-1,ix))
      enddo
    enddo
!$OMP END PARALLEL DO
  endsubroutine

  subroutine dz10th_op2(adj,add,model,data)
    logical,optional            :: add,adj
    real,dimension(nz,nx)     :: model,data

    c1=   35.0/294912.0
    c2= -405.0/229376.0
    c3=   567.0/40960.0
    c4=   -735.0/8192.0
    c5= 19845.0/16384.0
    model=0.0
!$OMP PARALLEL DO
    do ix=6,nx-5
      ! FREE SURFACE TERM
      model(1,ix) = model(1,ix) + 2.0*(c1*data(5,ix)+&
                                           c2*data(4,ix)+&
                                           c3*data(3,ix)+&
                                           c4*data(2,ix)+&
                                           c5*data(1,ix))
      model(2,ix) = model(2,ix) + c1*(data(6,ix)+data(4,ix))+&
                                      c2*(data(5,ix)+data(3,ix))+&
                                      c3*(data(4,ix)+data(2,ix))+&
                                      c4*(data(3,ix)+data(1,ix))+&
                                      c5*(data(2,ix)-data(1,ix))
      model(3,ix) = model(3,ix) + c1*(data(7,ix)+data(3,ix))+& 
                                      c2*(data(6,ix)+data(2,ix))+&
                                      c3*(data(5,ix)+data(1,ix))+&
                                      c4*(data(4,ix)-data(1,ix))+&
                                      c5*(data(3,ix)-data(2,ix))
      model(4,ix) = model(4,ix) + c1*(data(8,ix)+data(2,ix))+& 
                                      c2*(data(7,ix)+data(1,ix))+&
                                      c3*(data(6,ix)-data(1,ix))+&
                                      c4*(data(5,ix)-data(2,ix))+&
                                      c5*(data(4,ix)-data(3,ix))
      model(5,ix) = model(5,ix) + c1*(data(9,ix)+data(1,ix))+& 
                                      c2*(data(8,ix)-data(1,ix))+&
                                      c3*(data(7,ix)-data(2,ix))+&
                                      c4*(data(6,ix)-data(3,ix))+&
                                      c5*(data(5,ix)-data(4,ix))
      !==================
      do iz=6,nz-5
        model(iz,ix) = model(iz,ix) + c1*(data(iz+4,ix)-data(iz-5,ix))+&
                                          c2*(data(iz+3,ix)-data(iz-4,ix))+&
                                          c3*(data(iz+2,ix)-data(iz-3,ix))+&
                                          c4*(data(iz+1,ix)-data(iz-2,ix))+&
                                          c5*(data(iz  ,ix)-data(iz-1,ix))
      enddo
    enddo
!$OMP END PARALLEL DO
  endsubroutine

  subroutine dz4th_op4(adj,add,model,data)
    logical,optional            :: add,adj
    real,dimension(nz,nx)     :: model,data

    c1=-1.0/24.0
    c2= 9.0/8.0
    model=0.0
!$OMP PARALLEL DO
    do ix=3,nx-2
      do iz=3,nz-2
        model(iz,ix) = model(iz,ix) + c1*(data(iz+2,ix)-data(iz-1,ix))+&
                                          c2*(data(iz+1,ix)-data(iz  ,ix))
      enddo
    enddo
!$OMP END PARALLEL DO
  endsubroutine

  subroutine dz10th_op4(adj,add,model,data)
    logical,optional            :: add,adj
    real,dimension(nz,nx)     :: model,data

    c1=   35.0/294912.0
    c2= -405.0/229376.0
    c3=   567.0/40960.0
    c4=   -735.0/8192.0
    c5= 19845.0/16384.0
    model=0.0
!$OMP PARALLEL DO
    do ix=6,nx-5
      ! FREE SURFACE TERM
      model(1,ix) = model(1,ix) + c1*(data(6,ix)+data(4,ix))+&
                                      c2*(data(5,ix)+data(3,ix))+&
                                      c3*(data(4,ix)+data(2,ix))+&
                                      c4*(data(3,ix)+data(1,ix))+&
                                      c5*(data(2,ix)-data(1,ix))
      model(2,ix) = model(2,ix) + c1*(data(7,ix)+data(3,ix))+&
                                      c2*(data(6,ix)+data(2,ix))+&
                                      c3*(data(5,ix)+data(1,ix))+&
                                      c4*(data(4,ix)-data(1,ix))+&
                                      c5*(data(3,ix)-data(2,ix))
      model(3,ix) = model(3,ix) + c1*(data(8,ix)+data(2,ix))+&
                                      c2*(data(7,ix)+data(1,ix))+&
                                      c3*(data(6,ix)-data(1,ix))+&
                                      c4*(data(5,ix)-data(2,ix))+&
                                      c5*(data(4,ix)-data(3,ix))
      model(4,ix) = model(4,ix) + c1*(data(9,ix)+data(1,ix))+&
                                      c2*(data(8,ix)-data(1,ix))+&
                                      c3*(data(7,ix)-data(2,ix))+&
                                      c4*(data(6,ix)-data(3,ix))+&
                                      c5*(data(5,ix)-data(4,ix))
      !==================
      do iz=5,nz-5
        model(iz,ix) = model(iz,ix) + c1*(data(iz+5,ix)-data(iz-4,ix))+&
                                      c2*(data(iz+4,ix)-data(iz-3,ix))+&
                                      c3*(data(iz+3,ix)-data(iz-2,ix))+&
                                      c4*(data(iz+2,ix)-data(iz-1,ix))+&
                                      c5*(data(iz+1,ix)-data(iz  ,ix))
      enddo
    enddo
!$OMP END PARALLEL DO
  endsubroutine

  ! mixed derivative (shear stress) subroutines
  subroutine dxz4th_op2(adj,add,model,data1,data2)
    logical,optional            :: add,adj
    real,dimension(nz,nx)     :: model,data1,data2

    c1=-1.0/24.0
    c2= 9.0/8.0
    model=0.0
!$OMP PARALLEL DO
    do ix=3,nx-2
      do iz=3,nz-2
        !if(iz.lt.(nz-2))then
          model(iz,ix) = model(iz,ix) + c1*(data1(iz+1,ix)-data1(iz-2,ix))+&
                                            c2*(data1(iz  ,ix)-data1(iz-1,ix))+&
                                            c1*(data2(iz,ix+1)-data2(iz,ix-2))+&
                                            c2*(data2(iz,ix  )-data2(iz,ix-1))
        !else
        !  model(iz,ix) = model(iz,ix) +     data1(iz,ix) - data1(iz-1,ix) + &
        !                                c1*(data2(iz+2,ix)-data2(iz-2,ix))+&
        !                                c2*(data2(iz+1,ix)-data2(iz-1,ix))
          !if(ix.eq.3) then
          !  model(iz,3) = model(iz,3) + data1(iz,3) - data1(iz-1,3) + &
          !                              data2(iz,3) - data2(iz,2)
          !endif
        !endif
      enddo
    enddo
!$OMP END PARALLEL DO
  endsubroutine

  subroutine dxz10th_op2(adj,add,model,data1,data2)
    logical,optional            :: add,adj
    real,dimension(nz,nx)     :: model,data1, data2

    c1=   35.0/294912.0
    c2= -405.0/229376.0
    c3=   567.0/40960.0
    c4=   -735.0/8192.0
    c5= 19845.0/16384.0
    model=0.0
!$OMP PARALLEL DO
    do ix=6,nx-5
      ! FREE SURFACE TERM
      model(1,ix) = 0.0
      model(2,ix) = model(2,ix) + c1*(data1(6,ix)+data1(4,ix))+&
                                      c2*(data1(5,ix)+data1(3,ix))+&
                                      c3*(data1(4,ix)+data1(2,ix))+&
                                      c4*(data1(3,ix)+data1(1,ix))+&
                                      c5*(data1(2,ix)-data1(1,ix))+&
                                      c1*(data2(2,ix+4)-data2(2,ix-5))+&
                                      c2*(data2(2,ix+3)-data2(2,ix-4))+&
                                      c3*(data2(2,ix+2)-data2(2,ix-3))+&
                                      c4*(data2(2,ix+1)-data2(2,ix-2))+&
                                      c5*(data2(2,ix  )-data2(2,ix-1))
      model(3,ix) = model(3,ix) + c1*(data1(7,ix)+data1(3,ix))+& 
                                      c2*(data1(6,ix)+data1(2,ix))+&
                                      c3*(data1(5,ix)+data1(1,ix))+&
                                      c4*(data1(4,ix)-data1(1,ix))+&
                                      c5*(data1(3,ix)-data1(2,ix))+&
                                      c1*(data2(3,ix+4)-data2(3,ix-5))+&
                                      c2*(data2(3,ix+3)-data2(3,ix-4))+&
                                      c3*(data2(3,ix+2)-data2(3,ix-3))+&
                                      c4*(data2(3,ix+1)-data2(3,ix-2))+&
                                      c5*(data2(3,ix  )-data2(3,ix-1))
      model(4,ix) = model(4,ix) + c1*(data1(8,ix)+data1(2,ix))+& 
                                      c2*(data1(7,ix)+data1(1,ix))+&
                                      c3*(data1(6,ix)-data1(1,ix))+&
                                      c4*(data1(5,ix)-data1(2,ix))+&
                                      c5*(data1(4,ix)-data1(3,ix))+&
                                      c1*(data2(4,ix+4)-data2(4,ix-5))+&
                                      c2*(data2(4,ix+3)-data2(4,ix-4))+&
                                      c3*(data2(4,ix+2)-data2(4,ix-3))+&
                                      c4*(data2(4,ix+1)-data2(4,ix-2))+&
                                      c5*(data2(4,ix  )-data2(4,ix-1))
      model(5,ix) = model(5,ix) + c1*(data1(9,ix)+data1(1,ix))+& 
                                      c2*(data1(8,ix)-data1(1,ix))+&
                                      c3*(data1(7,ix)-data1(2,ix))+&
                                      c4*(data1(6,ix)-data1(3,ix))+&
                                      c5*(data1(5,ix)-data1(4,ix))+&
                                      c1*(data2(5,ix+4)-data2(5,ix-5))+&
                                      c2*(data2(5,ix+3)-data2(5,ix-4))+&
                                      c3*(data2(5,ix+2)-data2(5,ix-3))+&
                                      c4*(data2(5,ix+1)-data2(5,ix-2))+&
                                      c5*(data2(5,ix  )-data2(5,ix-1))
      !==================
      do iz=6,nz-5
        model(iz,ix) = model(iz,ix) + c1*(data1(iz+4,ix)-data1(iz-5,ix))+&
                                          c2*(data1(iz+3,ix)-data1(iz-4,ix))+&
                                          c3*(data1(iz+2,ix)-data1(iz-3,ix))+&
                                          c4*(data1(iz+1,ix)-data1(iz-2,ix))+&
                                          c5*(data1(iz  ,ix)-data1(iz-1,ix))+&
                                          c1*(data2(iz,ix+4)-data2(iz,ix-5))+&
                                          c2*(data2(iz,ix+3)-data2(iz,ix-4))+&
                                          c3*(data2(iz,ix+2)-data2(iz,ix-3))+&
                                          c4*(data2(iz,ix+1)-data2(iz,ix-2))+&
                                          c5*(data2(iz,ix  )-data2(iz,ix-1))
      enddo
    enddo
!$OMP END PARALLEL DO
  endsubroutine

endmodule
