! Program created by Gustavo Catao Alves at
! the Stanford Exploration Project
! May, 6th, 2014

! source module
module source_mod_new

  use sep

  implicit none
  real, private      :: fp, dt, dx
  integer, private   :: nsource, nx, nz

  contains

  ! Initialization of all variables needed by this subroutine
  subroutine source_init(fp_in,dt_in, dx_in, nsource_in, nz_in, nx_in)
    real                :: fp_in, dt_in, dx_in
    integer             :: nsource_in, nx_in, nz_in

    fp = fp_in
    dt = dt_in
    dx = dx_in
    nsource = nsource_in
    nz = nz_in
    nx = nx_in
  end subroutine

  ! Function that is called by elastic_mod
  function source_op(it, source) result(stat)
    real,intent(inout):: source
    integer,intent(in):: it
    integer           :: stat

    call source_op2(it, source)
    stat=0
  end function

  ! Ricker source
  subroutine source_op2(it, source)
    real               :: pi, aux, t
    integer            :: it
    real               :: source

    source=0.0
    pi = 3.14159265
    t = (it*dt)-(nsource*dt/2.0)
    aux = (pi*pi*fp*fp*t*t)
    source = (1.0 - 2.0*aux)*exp(-aux)
    if (it.le.(nsource/10)) source=0

  endsubroutine

  ! Distribution of point sources to create an explosive source
  subroutine explosive_source(source,xsource,zsource,sourcexx,sourcezz)
    real                    :: source,amp
    integer                 :: xsource,zsource
    real,dimension(nz,nx)   :: sourcexx,sourcezz

    !Scaling source energy to real air gun values, which are 14-21MPa
    !for a chamber volume of about 70cu, which in Newtons is 670-1015KN
    amp = 0.7E7 !Value tested for a constant solid: vp=2000.0 vs=1000.0 rho=1000.0
    source = amp*source*(dt/(dx*dx))*0.25
    sourcexx(zsource  ,xsource-1) = sourcexx(zsource  ,xsource-1) + source
    sourcexx(zsource  ,xsource  ) = sourcexx(zsource  ,xsource  ) + source
    sourcexx(zsource+1,xsource-1) = sourcexx(zsource+1,xsource-1) + source
    sourcexx(zsource+1,xsource  ) = sourcexx(zsource+1,xsource  ) + source
    sourcezz(zsource  ,xsource-1) = sourcezz(zsource  ,xsource-1) + source
    sourcezz(zsource  ,xsource  ) = sourcezz(zsource  ,xsource  ) + source
    sourcezz(zsource+1,xsource-1) = sourcezz(zsource+1,xsource-1) + source
    sourcezz(zsource+1,xsource  ) = sourcezz(zsource+1,xsource  ) + source

  endsubroutine

  ! Distribution of point sources to create a shear source
  subroutine shear_source(source,xsource,zsource,sourcexx,sourcezz)
    real                    :: source
    integer                 :: xsource,zsource
    real,dimension(nz,nx) :: sourcexx,sourcezz

    source = source*(dt/(dx*dx))*0.25
    sourcexx(zsource  ,xsource-1) = sourcexx(zsource  ,xsource-1) + source
    sourcexx(zsource  ,xsource  ) = sourcexx(zsource  ,xsource  ) + source
    sourcexx(zsource+1,xsource-1) = sourcexx(zsource+1,xsource-1) + source
    sourcexx(zsource+1,xsource  ) = sourcexx(zsource+1,xsource  ) + source
    sourcezz(zsource  ,xsource-1) = sourcezz(zsource  ,xsource-1) - source
    sourcezz(zsource  ,xsource  ) = sourcezz(zsource  ,xsource  ) - source
    sourcezz(zsource+1,xsource-1) = sourcezz(zsource+1,xsource-1) - source
    sourcezz(zsource+1,xsource  ) = sourcezz(zsource+1,xsource  ) - source

  endsubroutine

endmodule
