! Program created by Gustavo Catao Alves at
! the Stanford Exploration Project
! May, 6th, 2014

! finite difference module
module strain_mod

  use sep

  implicit none
  integer, private   :: nx, nz, iz, ix, nsnap, isnap
  real, private      :: c1,c2,c3,c4,c5

  contains

  ! Initialization of all variables needed by this subroutine
  subroutine strain_init(nz_in,nx_in,nsnap_in)
    integer        :: nx_in, nz_in, nsnap_in

    nx = nx_in
    nz = nz_in
    nsnap = nsnap_in

  endsubroutine

  ! Functions that are called by elastic_mod

  ! These x derivatives are centered at x=-1/2 
  function dx_op3(adj,add,model,data,isnap) result(stat)
    logical,intent(in),optional :: adj,add
    real,dimension(:,:)         :: model
    real,dimension(:,:,:)       :: data
    integer                     :: stat,isnap

    call dx10th_op4(adj,add,model,data,isnap)
    stat=0
  endfunction

  ! These z derivatives are centered at z=+1/2 
  function dz_op3(adj,add,model,data,isnap) result(stat)
    logical,intent(in),optional :: adj,add
    real,dimension(:,:)         :: model
    real,dimension(:,:,:)       :: data
    integer                     :: stat,isnap

    call dz10th_op4(adj,add,model,data,isnap)
    stat=0
  endfunction

  ! Derivative used on the shear stress
  function dxz_op(adj,add,model,data1,data2,isnap) result(stat)
    logical,intent(in),optional :: adj,add
    real,dimension(:,:)         :: model
    real,dimension(:,:,:)       :: data1,data2
    integer                     :: stat,isnap

    call dxz10th_op2(adj,add,model,data1,data2,isnap)
    stat=0
  endfunction

  ! Subroutines that are called by the functions
  subroutine dx10th_op4(adj,add,model,data,isnap)
    logical,optional            :: add,adj
    real,dimension(nz,nx)       :: model
    real,dimension(nz,nx,nsnap) :: data
    integer                     :: isnap

    c1=   35.0/294912.0
    c2= -405.0/229376.0
    c3=   567.0/40960.0
    c4=   -735.0/8192.0
    c5= 19845.0/16384.0
    model=0.0
!$OMP PARALLEL DO
    do ix=5,nx-5
      do iz=1,nz
        model(iz,ix) = model(iz,ix) + c1*(data(iz,ix+5,isnap)-data(iz,ix-4,isnap))+&
                                      c2*(data(iz,ix+4,isnap)-data(iz,ix-3,isnap))+&
                                      c3*(data(iz,ix+3,isnap)-data(iz,ix-2,isnap))+&
                                      c4*(data(iz,ix+2,isnap)-data(iz,ix-1,isnap))+&
                                      c5*(data(iz,ix+1,isnap)-data(iz,ix  ,isnap))
      enddo
    enddo
!$OMP END PARALLEL DO
  endsubroutine

  subroutine dz10th_op4(adj,add,model,data,isnap)
    logical,optional            :: add,adj
    real,dimension(nz,nx)       :: model
    real,dimension(nz,nx,nsnap) :: data
    integer                     :: isnap

    c1=   35.0/294912.0
    c2= -405.0/229376.0
    c3=   567.0/40960.0
    c4=   -735.0/8192.0
    c5= 19845.0/16384.0
    model=0.0
!$OMP PARALLEL DO
    do ix=1,nx
      do iz=5,nz-5
        model(iz,ix) = model(iz,ix) + c1*(data(iz+5,ix,isnap)-data(iz-4,ix,isnap))+&
                                      c2*(data(iz+4,ix,isnap)-data(iz-3,ix,isnap))+&
                                      c3*(data(iz+3,ix,isnap)-data(iz-2,ix,isnap))+&
                                      c4*(data(iz+2,ix,isnap)-data(iz-1,ix,isnap))+&
                                      c5*(data(iz+1,ix,isnap)-data(iz  ,ix,isnap))
      enddo
    enddo
!$OMP END PARALLEL DO
  endsubroutine

  subroutine dxz10th_op2(adj,add,model,data1,data2,isnap)
    logical,optional            :: add,adj
    real,dimension(nz,nx)       :: model
    real,dimension(nz,nx,nsnap) :: data1, data2
    integer                     :: isnap

    c1=   35.0/294912.0
    c2= -405.0/229376.0
    c3=   567.0/40960.0
    c4=   -735.0/8192.0
    c5= 19845.0/16384.0
    model=0.0
!$OMP PARALLEL DO
    do ix=6,nx-4
      do iz=6,nz-4
        model(iz,ix) = model(iz,ix) + c1*(data1(iz+4,ix,isnap)-data1(iz-5,ix,isnap))+&
                                      c2*(data1(iz+3,ix,isnap)-data1(iz-4,ix,isnap))+&
                                      c3*(data1(iz+2,ix,isnap)-data1(iz-3,ix,isnap))+&
                                      c4*(data1(iz+1,ix,isnap)-data1(iz-2,ix,isnap))+&
                                      c5*(data1(iz  ,ix,isnap)-data1(iz-1,ix,isnap))+&
                                      c1*(data2(iz,ix+4,isnap)-data2(iz,ix-5,isnap))+&
                                      c2*(data2(iz,ix+3,isnap)-data2(iz,ix-4,isnap))+&
                                      c3*(data2(iz,ix+2,isnap)-data2(iz,ix-3,isnap))+&
                                      c4*(data2(iz,ix+1,isnap)-data2(iz,ix-2,isnap))+&
                                      c5*(data2(iz,ix  ,isnap)-data2(iz,ix-1,isnap))
      enddo
    enddo
!$OMP END PARALLEL DO
  endsubroutine

endmodule
